
//////////////////////////////////////////////////////////////////////////////
//
//  xinout.C
//
//  $Log$
//
//	Written by Shaun Terry, Copyright 1995
//
//////////////////////////////////////////////////////////////////////////////

#ifndef lint
static const char rcsid[] = "$Id$";
#endif

#define XTFUNCPROTO

#include <X11/Intrinsic.h>
#include <Xm/Xm.h>

#include <Xm/Protocols.h>
#include <Xm/RowColumn.h>
#include <Xm/PushBG.h>
#include <Xm/Form.h>
#include <Xm/DrawingA.h>
#include <Xm/MessageB.h>
#include <Xm/DialogS.h>
#include <Xm/Label.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>
#include <Xm/PushBG.h>
#include <Xm/DialogS.h>
#include <Xm/Label.h>
#include <Xm/Text.h>

extern "C" XtAppContext _XtDefaultAppContext();
extern "C" int read(int, char *, int);

#include "xinout.h"
#include "socket.h"

#define ICONFILE	"/usr/local/images/xinout.icon"

#define MINIFONT	"*misc-fixed-medium-r-normal--9-*"
#define FORTUNEFONT	"*misc-fixed-medium-r-normal--10-*"
#define BIGFONT		"*misc-fixed-medium-r-normal--13-*"
#define BOLDFONT	"*adobe-times-bold-r-normal--14-*"
#define DETAILSFONT	"*adobe-helvetica-medium-r-normal--12-*"
#define DTLSHDRFONT	"*adobe-helvetica-bold-o-normal--14-*"

#define BOXWIDTH	26

char **Reasons;
int NumReasons;

PeopleInfo **People;
int NumPeople;

char *ProgTitle = "In/Out Board";
char *PersonalDescFile = ".xio-mydescs";
char *IconFile = ICONFILE;

int fFortuneShowing = 0;

#define MAXGRPMSG	4096

char GroupMsg[MAXGRPMSG];

struct {
    char *name;
	void (*callback)(Widget, caddr_t, caddr_t);
    caddr_t client_info;
} CommandItems[] = {
	{ "Set Group Message", cbSetGMsg, "S" },
	{ "Remove Group Message", cbSetGMsg, "R" },
	{ "Fortune Off", cbFortune, NULL },
	{ "About", cbAbout, NULL }
};

Pixmap in_pix, not_in_pix, out_pix, not_out_pix;
Pixmap arrow_pix;

int IOStat[ MAXPEOPLE ], SelectedReason=0, SelectedUser;
Widget SelectedWidget;
char Comment[MAXPEOPLE][MAXCOMM+1];

unsigned long ForeColor, BackColor;
unsigned long WhiteColor, BlackColor;
int CurrentUser=(-1);
Dimension MinWindowH;
Widget toplevel;

int fGMShowing = 0;
XmFontList grpmsgfont, fortunefont;

Widget widget_list[1024];

Widget commpopup, gmpopup, dtlspopup;

int fDebug;

Socket server;
XtInputId inputId=0;

const char *usage = "xinout [ -config file ] [ -name name { -in | -out | -comment text } | -fortune text | -grpmsg text ] [ -debug ] [ -help ]\n";

int main(int argc, char **argv)
{
    int fCmdLineOnly = 0, fIn=0, fOut=0;
	char *who = 0, *pcomm = 0, *grpmsg = 0, *fortune = 0;
	char *cfile = "/usr/local/xinout.cfg";

	if (getParamFlag(argc, argv, "-help")) {
		printf(usage);
		return 0;
	}

	fDebug = getParamFlag(argc, argv, "-debug");
    getParamString(argc, argv, "-config", cfile);
    getParamString(argc, argv, "-name", who);

    if ((fIn = getParamFlag(argc, argv, "-in"))) fCmdLineOnly = 1;
    if ((fOut = getParamFlag(argc, argv, "-out"))) fCmdLineOnly = 1;
    if (getParamString(argc, argv, "-comment", pcomm)) fCmdLineOnly = 1;
    if (getParamString(argc, argv, "-grpmsg", grpmsg)) fCmdLineOnly = 1;
    if (getParamString(argc, argv, "-fortune", fortune)) fCmdLineOnly = 1;

	if (!who && fCmdLineOnly) {
		printf(usage);
		return(1);
	}

    for (int i=0; i < MAXPEOPLE; ++i) { IOStat[i] = 0; Comment[i][0] = '\0'; }

    if (fCmdLineOnly) {
		if (fIn) sendToServer(STATUS_CHANGE, who, 'I');
		if (fOut) sendToServer(STATUS_CHANGE, who, 'O');
		if (pcomm) sendToServer(COMMENT_CHANGE, who, 'x', pcomm);
		if (grpmsg) sendToServer(SET_GROUP_MSG, "", 'x', grpmsg);
		if (fortune) sendToServer(SET_FORTUNE, "", 'x', fortune);
		return(0);
    }

	readConfigFile(cfile);

    toplevel = XtInitialize("xinout", "Xinout", NULL, 0,
#if XtSpecificationRelease == 4
									(Cardinal *) &argc, argv);
#else
									&argc, argv);
#endif

    createWidgets(toplevel);

	connectToServer();

    XtMainLoop();
}

void cbEndit()
{
    FreePixmaps();
    exit(0);
}

void createWidgets(Widget toplevel)
{
    Arg arg[32];
    Widget w, form, button, mainForm;
    XmString xms;

    mainForm = XtCreateManagedWidget("", xmFormWidgetClass, toplevel, NULL, 0);

    Display *dpy = XtDisplay(toplevel);

    WhiteColor = WhitePixel(dpy, DefaultScreen(dpy));
    BlackColor = BlackPixel(dpy, DefaultScreen(dpy));

    XmFontList xfs = getFontList(dpy, MINIFONT);
    grpmsgfont = getFontList(dpy, BIGFONT);
    XmFontList xfs_b = getFontList(dpy, BOLDFONT);
    fortunefont = getFontList(dpy, FORTUNEFONT);
    XmFontList xfs_dtls = getFontList(dpy, DETAILSFONT);
    XmFontList xfs_dtlshdr = getFontList(dpy, DTLSHDRFONT);

    Widget rc = XtVaCreateManagedWidget("", xmRowColumnWidgetClass, mainForm,
		XmNspacing, 0,
		XmNmarginWidth, 0,
		XmNmarginHeight, 0,
		NULL);

    XmString xmsblank = stringCreate(" ");
    w = XtVaCreateManagedWidget("who", xmLabelWidgetClass, rc,
		XmNborderWidth, 1,
		XmNfontList, xfs,
		XmNlabelString, xmsblank,
		NULL);
    XtAddEventHandler(w, ButtonPressMask, False, (XtEventHandler) evMenuPopup, NULL);
    widget_list[WHOCOL + NumPeople] = w;

    for (int i=0; i < NumPeople; ++i) {

	xms = stringCreate(People[i]->Name);
	w = XtVaCreateManagedWidget("who", xmLabelWidgetClass, rc,
		XmNborderWidth, 1,
		XmNfontList, xfs,
		XmNlabelString, xms,
		XmNalignment, XmALIGNMENT_BEGINNING,
		XmNmarginLeft, 2,
		NULL);
	stringFree(xms);
	XtAddEventHandler(w, ButtonPressMask, False, (XtEventHandler) evPerson, (caddr_t)i);

	widget_list[WHOCOL + i] = w;
    }

    rc = XtVaCreateManagedWidget("", xmRowColumnWidgetClass, mainForm,
		XmNspacing, 0,
		XmNpacking, XmPACK_COLUMN,
		XmNmarginWidth, 0,
		XmNmarginHeight, 0,
		XmNisAligned, False,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, rc,
		NULL);

    xms = stringCreate("IN");
    w = XtVaCreateManagedWidget("in", xmLabelWidgetClass, rc,
		XmNborderWidth, 1,
		XmNfontList, xfs,
		XmNwidth, BOXWIDTH,
		XmNrecomputeSize, False,
		XmNlabelString, xms,
		NULL);
    stringFree(xms);

    Dimension BoxHeight;
    XtVaGetValues(w, XmNheight, &BoxHeight, NULL);

    for (i=0; i < NumPeople; ++i) {
	w = XtVaCreateManagedWidget("in", xmDrawingAreaWidgetClass, rc,
		XmNborderWidth, 1,
		XmNwidth, BOXWIDTH,
		XmNheight, BoxHeight,
		XmNmarginHeight, 0,
		XmNmarginWidth, 0,
		XmNresizePolicy, XmRESIZE_NONE,
		NULL);
	XtAddCallback(w, XmNinputCallback, (XtCallbackProc) cbIn, (caddr_t) i);

	widget_list[INCOL + i] = w;
    }

    rc = XtVaCreateManagedWidget("", xmRowColumnWidgetClass, mainForm,
		XmNspacing, 0,
		XmNmarginWidth, 0,
		XmNpacking, XmPACK_COLUMN,
		XmNmarginHeight, 0,
		XmNisAligned, False,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, rc,
		NULL);

    xms = stringCreate("OUT");
    w = XtVaCreateManagedWidget("out", xmLabelWidgetClass, rc,
		XmNborderWidth, 1,
		XmNfontList, xfs,
		XmNwidth, BOXWIDTH,
		XmNrecomputeSize, False,
		XmNlabelString, xms,
		NULL);
    stringFree(xms);

    for (i=0; i < NumPeople; ++i) {
	w = XtVaCreateManagedWidget("out", xmDrawingAreaWidgetClass, rc,
		XmNborderWidth, 1,
		XmNwidth, BOXWIDTH,
		XmNheight, BoxHeight,
		XmNmarginHeight, 0,
		XmNmarginWidth, 0,
		XmNresizePolicy, XmRESIZE_NONE,
		NULL);
	XtAddCallback(w, XmNinputCallback, (XtCallbackProc) cbOut, (caddr_t) i);

	widget_list[OUTCOL + i] = w;
    }

    rc = XtVaCreateManagedWidget("", xmRowColumnWidgetClass, mainForm,
		XmNspacing, 0,
		XmNmarginWidth, 0,
		XmNmarginHeight, 0,
		XmNisAligned, False,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, rc,
		XmNrightAttachment, XmATTACH_FORM,
		NULL);

    xms = stringCreate("Comments");
    w = XtVaCreateManagedWidget("comms", xmLabelWidgetClass, rc,
		XmNborderWidth, 1,
		XmNfontList, xfs,
		XmNlabelString, xms,
		XmNrecomputeSize, False,
		NULL);
    widget_list[COMMSCOL + NumPeople] = w;
    stringFree(xms);

    for (i=0; i < NumPeople; ++i) {
	w = XtVaCreateManagedWidget("comms", xmLabelWidgetClass, rc,
		XmNborderWidth, 1,
		XmNfontList, xfs,
		XmNlabelString, xmsblank,
		XmNwidth, 100,
		XmNalignment, XmALIGNMENT_BEGINNING,
		XmNmarginLeft, 2,
		XmNrecomputeSize, False,
		NULL);
	XtAddEventHandler(w, ButtonPressMask, False, (XtEventHandler) evComment, (caddr_t) i);

	widget_list[COMMSCOL + i] = w;
    }

    XtVaGetValues(widget_list[WHOCOL+0],
		XmNforeground, &ForeColor, XmNbackground, &BackColor,
		NULL);

    w = XtVaCreateWidget("GroupMessage", xmLabelWidgetClass, mainForm,
		XmNrightAttachment, XmATTACH_FORM,
		XmNleftAttachment, XmATTACH_FORM,
		XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, rc,
		XmNfontList, grpmsgfont,
		XmNrecomputeSize, True,
		XmNmarginHeight, 10,
		NULL);
    widget_list[GROUPMSG] = w;

    stringFree(xmsblank);

    /*======================= COMMAND MENU POPUP =========================*/

    rc = XmCreatePopupMenu(toplevel, "Command", NULL, 0);

    widget_list[COMMANDPOPUP] = rc;

    for (i=0; i < XtNumber(CommandItems); ++i) { 
		w = XtVaCreateManagedWidget(CommandItems[i].name,
				xmPushButtonGadgetClass, rc,
				XmNfontList, xfs_b, NULL);
		XtAddCallback(w, XmNactivateCallback,
			(XtCallbackProc) CommandItems[i].callback, CommandItems[i].client_info);
    }

    /*======================= GROUP MESSAGE POPUP =========================*/

    static PopupButtonInfo gmpbi[] = {
	{ "Apply",   (XtCallbackProc) cbGMApply, NULL },
	{ "Dismiss", NULL, NULL },
	{ NULL }
    };

    gmpopup = createPopupWindow("Set Group Message", toplevel, gmpbi,
														&form, &button);

    w = XtVaCreateManagedWidget("", xmTextWidgetClass, form,
		XmNtopAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_WIDGET,
		XmNbottomWidget, button,
    	XmNeditMode, XmSINGLE_LINE_EDIT,
		XmNfontList, xfs_b,
		XmNwidth, 200,
		NULL);
    widget_list[GMSGFLD] = w;

    /*======================== DESCRIPTION POPUP ==========================*/

    static PopupButtonInfo commpbi[] = {
		{ "Apply",   (XtCallbackProc) cbCommentApply, NULL },
		{ "Dismiss", NULL, NULL },
		{ NULL }
    };

    commpopup = createPopupWindow("Comments", toplevel, commpbi,&form,&button);

    w = XtVaCreateManagedWidget("", xmLabelWidgetClass, form,
		XmNleftAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		XmNalignment, XmALIGNMENT_CENTER,
		XmNmarginHeight, 5,
		NULL);

    widget_list[COMMENTLABEL] = w;

    rc = XtVaCreateManagedWidget("", xmRowColumnWidgetClass, form,
		XmNpacking, XmPACK_COLUMN,
		XmNmarginWidth, 0,
		XmNmarginHeight, 0,
		XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, w,
		XmNleftAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		XmNspacing, 0,
		XmNisAligned, False,
		NULL);

    for (i=0; i < NumReasons; ++i) {
		xms = stringCreate(Reasons[i]);
		w = XtVaCreateManagedWidget("Reasons", xmPushButtonGadgetClass, rc,
												XmNlabelString, xms, NULL);
		stringFree(xms);
		XtAddCallback(w, XmNactivateCallback, (XtCallbackProc) cbReasons, (caddr_t) i);
    }

    {
	char fname[256], *path;
	char rsn[256];
	FILE *fp;

	fname[0] = '\0'; 
	if ((path = getenv("HOME"))) sprintf(fname, "%s/", path);
	strcat(fname, PersonalDescFile);
	if ((fp = fopen(fname, "r")) != NULL) {
	    while (fgets(rsn, 127, fp) != NULL) {
			if (rsn[0] == ' ' || rsn[0] == '\t' || rsn[0] == '\n'
						|| rsn[0] == '\0') continue;
			rsn[ strlen(rsn) - 1 ] = '\0';
			if (strlen(rsn) > MAXCOMM) rsn[MAXCOMM] = '\0';
			if (strlen(rsn) == 0) continue;
			xms = stringCreate(rsn);
			w = XtVaCreateManagedWidget("Reasons", xmPushButtonGadgetClass,
				rc, XmNlabelString, xms,
				NULL);
			stringFree(xms);
			XtAddCallback(w, XmNactivateCallback, (XtCallbackProc) cbReasons, (caddr_t) i);
	    }
	    fclose(fp);
	}
    }

    w = XtVaCreateManagedWidget("", xmTextWidgetClass, form,
		XmNleftAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_WIDGET,
		XmNbottomWidget, button,
		XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, rc,
    		XmNeditMode, XmSINGLE_LINE_EDIT,
		XmNfontList, xfs_b,
		XmNwidth, 100,
		NULL);
    widget_list[COMMENTTEXT] = w;

    /*======================== DETAILS POPUP ==========================*/

    static PopupButtonInfo dtlspbi[] = {
	{ "Dismiss", NULL, NULL },
	{ NULL }
    };

    dtlspopup = createPopupWindow("Details", toplevel, dtlspbi,&form,&button);

    w = XtVaCreateManagedWidget("", xmLabelWidgetClass, form,
		XmNleftAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		XmNfontList, xfs_dtlshdr,
		NULL);
    widget_list[DTLS_NAME] = w;

    rc = XtVaCreateManagedWidget("", xmRowColumnWidgetClass, form,
		XmNpacking, XmPACK_COLUMN,
		XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, w,
		XmNbottomAttachment, XmATTACH_WIDGET,
		XmNbottomWidget, button,
		NULL);

    XtSetArg(arg[0], XmNfontList, xfs_dtls);
    XtCreateManagedWidget("Office:", xmLabelWidgetClass, rc, arg, 1);
    XtCreateManagedWidget("Home:", xmLabelWidgetClass, rc, arg, 1);
    XtCreateManagedWidget("Beeper:", xmLabelWidgetClass, rc, arg, 1);
    XtCreateManagedWidget("E-mail:", xmLabelWidgetClass, rc, arg, 1);

    rc = XtVaCreateManagedWidget("", xmRowColumnWidgetClass, form,
		XmNpacking, XmPACK_COLUMN,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, rc,
		XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, w,
		XmNrightAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_WIDGET,
		XmNbottomWidget, button,
		NULL);

    XtSetArg(arg[0], XmNalignment, XmALIGNMENT_BEGINNING);
    XtSetArg(arg[1], XmNfontList, xfs_dtls);
    w = XtCreateManagedWidget("", xmLabelWidgetClass, rc, arg, 2);
    widget_list[DTLS_OPHONE] = w;
    w = XtCreateManagedWidget("", xmLabelWidgetClass, rc, arg, 2);
    widget_list[DTLS_HPHONE] = w;
    w = XtCreateManagedWidget("", xmLabelWidgetClass, rc, arg, 2);
    widget_list[DTLS_BEEPNUM] = w;
    w = XtCreateManagedWidget("", xmLabelWidgetClass, rc, arg, 2);
    widget_list[DTLS_EMAIL] = w;

    /*--------------------------------------------------------------------*/

    XtSetMappedWhenManaged(toplevel, False);

#ifdef XXX
    basew->setIconToPixmapFile(IconFile);
#endif

    XtRealizeWidget(toplevel);

	XmAddWMProtocolCallback(toplevel,
		XmInternAtom(dpy, "WM_DELETE_WINDOW", 0),
		(XtCallbackProc) cbEndit, NULL);

    restrictSize(SetToCurrent, SetToCurrent, SetToCurrent, SetToUnrestricted);

    XtVaGetValues(toplevel, XmNheight, &MinWindowH, NULL);

    setTitle(toplevel, ProgTitle);

    XtMapWidget(toplevel);

    setIconName(toplevel, "Xinout");

    CreateBulletPixmap(widget_list[INCOL+0], &in_pix);
    CreateBulletPixmap(widget_list[OUTCOL+0], &out_pix);
    CreateBlankPixmap(widget_list[INCOL+0], &not_in_pix);
    CreateBlankPixmap(widget_list[OUTCOL+0], &not_out_pix);
    CreateArrowPixmap(widget_list[WHOCOL+0], &arrow_pix);

    XtVaSetValues(widget_list[WHOCOL+NumPeople],
		XmNlabelType, XmPIXMAP, XmNlabelPixmap, arrow_pix,
		XmNmarginHeight, 0,
		NULL);
}

/*======================================================================*/
/*			PIXMAP SECTION					*/
/*======================================================================*/

void CreateBulletPixmap(Widget widg, Pixmap *pPix)
{
    Pixmap pixmap;
    Display *dpy;
    unsigned long ForeColor, BackColor;
    Dimension w, h;
    XGCValues gcv;
    GC gc;

    XtVaGetValues(widg, 
		XmNwidth, (XtArgVal) &w,
		XmNheight, (XtArgVal) &h,
		XmNforeground, (XtArgVal) &ForeColor,
		XmNbackground, (XtArgVal) &BackColor,
		NULL);

    dpy = XtDisplay(widg);

    pixmap = XCreatePixmap(dpy, XtWindow(widg),
		w, h, DefaultDepth(dpy, DefaultScreen(dpy)));

    gcv.foreground = BackColor;
    gc = XCreateGC(dpy, pixmap, GCForeground, &gcv);

    XFillRectangle(dpy, pixmap, gc, 0, 0, w, h); /* Clear out pixmap */

    XFreeGC(dpy, gc);

    gcv.foreground = ForeColor;
    gcv.background = BackColor;
    gc = XCreateGC(dpy, pixmap, GCForeground|GCBackground, &gcv);

	int d = h/3;
	int c = (int) ((double(w) - double(d)) / 2.0);
    XDrawArc(dpy, pixmap, gc, c, d, d, d, 0, 360*64);
    XFillArc(dpy, pixmap, gc, c, d, d, d, 0, 360*64);

    XFreeGC(dpy, gc);

    *pPix = pixmap;
}

void CreateSquarePixmap(Widget widg, Pixmap *pPix)
{
    Pixmap pixmap;
    Display *dpy;
    unsigned long ForeColor, BackColor;
    Dimension w, h;
    XGCValues gcv;
    GC gc;

    XtVaGetValues(widg, 
		XmNwidth, (XtArgVal) &w,
		XmNheight, (XtArgVal) &h,
		XmNforeground, (XtArgVal) &ForeColor,
		XmNbackground, (XtArgVal) &BackColor,
		NULL);

    dpy = XtDisplay(widg);

    pixmap = XCreatePixmap(dpy, XtWindow(widg),
		w, h, DefaultDepth(dpy, DefaultScreen(dpy)));

    gcv.foreground = BackColor;
    gc = XCreateGC(dpy, pixmap, GCForeground, &gcv);

    XFillRectangle(dpy, pixmap, gc, 0, 0, w, h); /* Clear out pixmap */

    XFreeGC(dpy, gc);

    gcv.foreground = ForeColor;
    gcv.background = BackColor;
    gc = XCreateGC(dpy, pixmap, GCForeground|GCBackground, &gcv);

    XDrawRectangle(dpy, pixmap, gc, (w - (double(h)/3.0)) / 2.0, h/3, h/3, h/3);
    XFillRectangle(dpy, pixmap, gc, (w - (double(h)/3.0)) / 2.0, h/3, h/3, h/3);

    XFreeGC(dpy, gc);

    *pPix = pixmap;
}

void CreateBlankPixmap(Widget widg, Pixmap *pPix)
{
    Pixmap pixmap;
    Display *dpy;
    unsigned long BackColor;
    Dimension w, h;
    XGCValues gcv;
    GC gc;

    XtVaGetValues(widg, 
		XmNwidth, (XtArgVal) &w,
		XmNheight, (XtArgVal) &h,
		XmNbackground, (XtArgVal) &BackColor,
		NULL);

    dpy = XtDisplay(widg);

    pixmap = XCreatePixmap(dpy, XtWindow(widg),
		w, h, DefaultDepth(dpy, DefaultScreen(dpy)));

    gcv.foreground = BackColor;
    gc = XCreateGC(dpy, pixmap, GCForeground, &gcv);

    XFillRectangle(dpy, pixmap, gc, 0, 0, w, h); /* Clear out pixmap */

    XFreeGC(dpy, gc);

    *pPix = pixmap;
}

void CreateArrowPixmap(Widget widg, Pixmap *pPix)
{
    Pixmap pixmap;
    Display *dpy;
    unsigned long ForeColor, BackColor, i;
    Dimension w, h;
    XGCValues gcv;
    XPoint vertex[4];
    GC gc;
    static int lp[] = { 1, 4 };

    XtVaGetValues(widg, 
		XmNwidth, (XtArgVal) &w,
		XmNheight, (XtArgVal) &h,
		XmNforeground, (XtArgVal) &ForeColor,
		XmNbackground, (XtArgVal) &BackColor,
		NULL);

    dpy = XtDisplay(widg);

    pixmap = XCreatePixmap(dpy, XtWindow(widg),
		w, h, DefaultDepth(dpy, DefaultScreen(dpy)));

    gcv.foreground = BackColor;
    gc = XCreateGC(dpy, pixmap, GCForeground, &gcv);

    XFillRectangle(dpy, pixmap, gc, 0, 0, w, h); /* Clear out pixmap */

    XFreeGC(dpy, gc);

    gcv.foreground = ForeColor;
    gcv.background = BackColor;
    gc = XCreateGC(dpy, pixmap, GCForeground|GCBackground, &gcv);

    XDrawRectangle(dpy, pixmap, gc, w/3, h/3, (w/2) - (w/3), h/3);
    XFillRectangle(dpy, pixmap, gc, w/3, h/3, (w/2) - (w/3), h/3);
    vertex[0].x = w/2;
    vertex[0].y = (h/6);
    vertex[3].x = w/2;
    vertex[3].y = (h/6);
    vertex[1].x = w - (w/3) - 1;
    vertex[1].y = h/2;
    vertex[2].x = w/2;
    vertex[2].y = h - (h/6) - 1;
    XDrawLines(dpy, pixmap, gc, vertex, 4, CoordModeOrigin);
    XFillPolygon(dpy, pixmap, gc, vertex, 4, Convex, CoordModeOrigin);

    XFreeGC(dpy, gc);

    gcv.foreground = BackColor;
    gc = XCreateGC(dpy, pixmap, GCForeground, &gcv);

    for (i=0; i < XtNumber(lp); ++i) {
	XDrawLine(dpy, pixmap, gc, (w/3) + lp[i], 0, (w/3) + lp[i], h-1);
    }

    XFreeGC(dpy, gc);

    *pPix = pixmap;
}

void FreePixmaps()
{
    Display *dpy = XtDisplay(toplevel);

    XFreePixmap(dpy, in_pix);
    XFreePixmap(dpy, out_pix);
    XFreePixmap(dpy, not_in_pix);
    XFreePixmap(dpy, not_out_pix);
    XFreePixmap(dpy, arrow_pix);
}

#include <Xm/PushB.h>

Widget createPopupWindow(char *name, Widget parent, PopupButtonInfo *bi,
								Widget *form, Widget *button)
{
    Arg arg[32];
    int cnt = 0, i = 0;

	Widget pop =
		XtVaCreatePopupShell(name, xmDialogShellWidgetClass, parent, NULL);

    // Count the number of buttons specified
    PopupButtonInfo *tmpbi = bi;
    int nbuttons=0;
    if (tmpbi != NULL) {
		while (tmpbi->name != NULL) { ++nbuttons; ++tmpbi; }
    }

    cnt=0;
    XtSetArg(arg[cnt], XmNfractionBase, nbuttons+2); ++cnt;
    XtSetArg(arg[cnt], XmNallowOverlap, False); ++cnt;
    *form = XmCreateForm(pop, "", arg, cnt);

    if (bi != NULL) {

		Widget w;
		while (bi->name != NULL) {

			cnt=0;
			XtSetArg(arg[cnt], XmNbottomAttachment, XmATTACH_FORM); ++cnt;
			XtSetArg(arg[cnt], XmNleftAttachment, XmATTACH_POSITION); ++cnt;
			XtSetArg(arg[cnt], XmNleftPosition, i+1); ++cnt;
			XtSetArg(arg[cnt], XmNrightAttachment, XmATTACH_POSITION); ++cnt;
			XtSetArg(arg[cnt], XmNrightPosition, i+2); ++cnt;
			XtManageChild(w = XmCreatePushButton(*form, bi->name, arg, cnt));
			if (bi->callback != NULL)
				XtAddCallback(w, XmNactivateCallback, bi->callback,
				(caddr_t) bi->clientData);

			*button = w;

			++bi;
			++i;
    	}

		--bi;

		// If the callback for the last button created is null then a
		// a default behavior of popping the window down is assumed.
		if (bi->callback == NULL)
		XtAddCallback(w, XmNactivateCallback, cbPopupDismiss, (caddr_t) pop);
    }

/*
    if (clientWC != NULL) {
	cnt=0;
	XtSetArg(arg[cnt], XmNtopAttachment, XmATTACH_FORM); ++cnt;
	XtSetArg(arg[cnt], XmNleftAttachment, XmATTACH_FORM); ++cnt;
	XtSetArg(arg[cnt], XmNrightAttachment, XmATTACH_FORM); ++cnt;
	if (i > 0) {
	XtSetArg(arg[cnt], XmNbottomAttachment, XmATTACH_WIDGET); ++cnt;
	XtSetArg(arg[cnt], XmNbottomWidget, buttonWidg[0]); ++cnt;
	}
	else { XtSetArg(arg[cnt], XmNbottomAttachment, XmATTACH_FORM); ++cnt; }
	clientWidg = XtCreateManagedWidget("", clientWC, formWidg, arg, cnt);
    }
    else clientWidg = formWidg;
*/

	return *form;
}

void cbPopupDismiss(Widget, XtPointer client, XtPointer)
{
	XtUnmanageChild((Widget) client);
}

void readConfigFile(char *cfile)
{
    ConfigFile cf(cfile);

	if (!cf.exists()) {
		fprintf(stderr, "Config file \"%s\" does not exist.\n", cf.name());
		exit(1);
	}

    if (!cf.seekSection("DefaultComments")) {
		fprintf(stderr, "xinout: Cannot find any default comments!\n");
		exit(1);
	}

	NumReasons = 0;
    while (cf.readNext()) ++NumReasons;

	Reasons = (char **) new char[NumReasons * sizeof(char *)];

    cf.seekSection("DefaultComments");

	int i=0;
	char *p;
	while (p = cf.readNext()) {
			strcpy((Reasons[i++] = new char[strlen(p)+1]), p);
	}

    if (!cf.seekSection("Users")) {
		fprintf(stderr, "xinout: Cannot find any users!\n");
		exit(1);
	}

	NumPeople = 0;
    while (cf.readNext()) ++NumPeople;

	People = (PeopleInfo **) new char[NumPeople * sizeof(PeopleInfo *)];

    cf.seekSection("Users");

	i=0;
	while (p = cf.readNext()) {
			People[i] = new PeopleInfo;
			strcpy((People[i]->Name = new char[strlen(p)+1]), p);
    		People[i]->FullName = "";
		    People[i]->OfficePhone = "";
    		People[i]->HomePhone = "";
   			People[i]->BeeperNum = "";
    		People[i]->Email = "";
			++i;
	}

	for (i=0; i < NumPeople; ++i) {
		if (cf.seekSection(People[i]->Name)) {
			char *p;
			if (!(p = cf.readNext())) continue;
			People[i]->FullName = strdup(p);
			if (!(p = cf.readNext())) continue;
			People[i]->OfficePhone = strdup(p);
			if (!(p = cf.readNext())) continue;
			People[i]->HomePhone = strdup(p);
			if (!(p = cf.readNext())) continue;
			People[i]->BeeperNum = strdup(p);
			if (!(p = cf.readNext())) continue;
			People[i]->Email = strdup(p);
		}
	}

	ProgTitle = strdup(cf.readString("Title", ProgTitle));
	PersonalDescFile =
			strdup(cf.readString("PersonalCommentsFile", PersonalDescFile));
	IconFile = strdup(cf.readString("IconFile", IconFile));
}

void evComment(Widget w, caddr_t client, XEvent *ev, char *)
{
    XtUngrabPointer(w, CurrentTime);

    char p[64];
    XButtonPressedEvent *but_event = (XButtonPressedEvent *) ev;

    SelectedUser = (int) client;

    if (isButton1(but_event)) {

		sprintf(p, "Comments for %s", People[SelectedUser]->Name);

		setLabelString(COMMENTLABEL, p);

		XmTextSetString(widget_list[COMMENTTEXT], "");

		XtManageChild(commpopup);
    }
    else if (isButton2(but_event)) {

		XmTextSetString(widget_list[COMMENTTEXT], Comment[SelectedUser]);

		XtManageChild(commpopup);
    }
    else if (isButton3(but_event)) {
		change_comment(SelectedUser, "", TRUE);
    }
}

void change_comment(int forwho, char *txt, bool andSend)
{
    if (strlen(txt) > MAXCOMM) *(txt + MAXCOMM) = '\0';

    setLabelString(COMMSCOL+forwho, txt);

    // We do this little kludge to get around an internal Motif bug which
    // sometimes causes the new string to be aligned improperly vertically.
    Dimension h;
    XtVaGetValues(widget_list[COMMSCOL+forwho], XmNheight, &h, NULL);
    XtVaSetValues(widget_list[COMMSCOL+forwho], XmNheight, h, NULL);

    strcpy(Comment[forwho], txt);

	if (andSend) sendToServer(COMMENT_CHANGE, People[forwho]->Name, 'x', txt);
}

void cbCommentApply(Widget, caddr_t, caddr_t)
{
    char *txt;
    Boolean dofree = False;

    txt = XmTextGetString(widget_list[COMMENTTEXT]);

    if (txt == NULL) goto nxt;

    if (strlen(txt) == 0) {
	XtFree(txt);
nxt:
	if (SelectedReason >= NumReasons) {
	    txt = getLabelString(SelectedWidget);
	    dofree = True;
	} 
	else txt = Reasons[ SelectedReason ];
    }
    else dofree = True;

    change_comment(SelectedUser, txt, TRUE);

    if (dofree) XtFree(txt);

	XtUnmanageChild(commpopup);
}

void evPerson(Widget w, caddr_t client, XEvent *ev, char *)
{
    XtUngrabPointer(w, CurrentTime);

    XButtonPressedEvent *but_event = (XButtonPressedEvent *) ev;

    if (isButton1(but_event)) {
		display_details( (int) client );
		return;
    }

#ifdef XXX
    if ((int) client == CurrentUser) return;

    change_bprkpr((int) client);
//    bprkpr_change((int) client);
#endif
}

void display_details(int forwho)
{
    setLabelString(DTLS_NAME, People[forwho]->FullName);
    setLabelString(DTLS_OPHONE, People[forwho]->OfficePhone);
    setLabelString(DTLS_HPHONE, People[forwho]->HomePhone);
    setLabelString(DTLS_BEEPNUM, People[forwho]->BeeperNum);
    setLabelString(DTLS_EMAIL, People[forwho]->Email);

	XtManageChild(dtlspopup);
}

void setLabelString(int idx, char *str)
{
    XmString xms = XmStringCreateLtoR(str, XmSTRING_DEFAULT_CHARSET);
    XtVaSetValues(widget_list[idx], XmNlabelString, xms, NULL);
    XmStringFree(xms);
}

void setLabelString(Widget w, char *str)
{
    XmString xms = XmStringCreateLtoR(str, XmSTRING_DEFAULT_CHARSET);
    XtVaSetValues(w, XmNlabelString, xms, NULL);
    XmStringFree(xms);
}

void setMessageString(Widget w, char *str)
{
    XmString xms = XmStringCreateLtoR(str, XmSTRING_DEFAULT_CHARSET);
    XtVaSetValues(w, XmNmessageString, xms, NULL);
    XmStringFree(xms);
}

char *getLabelString(int idx)
{
    char *txt;
    XmString xms;

    XtVaGetValues(widget_list[idx], XmNlabelString, &xms, NULL);
    XmStringGetLtoR(xms, XmSTRING_DEFAULT_CHARSET, &txt);
    XmStringFree(xms);

    return txt;
}

char *getLabelString(Widget w)
{
    char *txt;
    XmString xms;

    XtVaGetValues(w, XmNlabelString, &xms, NULL);
    XmStringGetLtoR(xms, XmSTRING_DEFAULT_CHARSET, &txt);
    XmStringFree(xms);

    return txt;
}

void change_bprkpr(int towho)
{
    if (CurrentUser >= 0) {
	XtVaSetValues(widget_list[WHOCOL+CurrentUser],
		XmNforeground, ForeColor,
		XmNbackground, BackColor,
		NULL);
    }

    XtVaSetValues(widget_list[WHOCOL+towho],
		XmNforeground,
			(WhiteColor == ForeColor) ? BackColor : WhiteColor,
		XmNbackground, ForeColor,
		NULL);

    CurrentUser = towho;
}

void cbIn(Widget w, caddr_t client, caddr_t call)
{
    XtUngrabPointer(w, CurrentTime);

    XmDrawingAreaCallbackStruct *p = (XmDrawingAreaCallbackStruct *) call;

    if (p->reason != XmCR_INPUT || p->event->type != ButtonPress) return;

    int i = (int) client;

    if (IOStat[i] & IN_FLAG) return;

    change_status(i, 'I', TRUE);
}

void cbOut(Widget w, caddr_t client, caddr_t call)
{
    XtUngrabPointer(w, CurrentTime);

    XmDrawingAreaCallbackStruct *p = (XmDrawingAreaCallbackStruct *) call;

    if (p->reason != XmCR_INPUT || p->event->type != ButtonPress) return;

    int i = (int) client;

    if (IOStat[i] & OUT_FLAG) return;

    change_status(i, 'O', TRUE);
}

void change_status(int forwho, char in_or_out, bool andSend)
{
    Widget w1, w2;

	w1=(in_or_out=='I') ? widget_list[INCOL+forwho]:widget_list[OUTCOL+forwho];
	w2=(in_or_out=='I') ? widget_list[OUTCOL+forwho]:widget_list[INCOL+forwho];

	XtVaSetValues(w1,
		XmNbackgroundPixmap, ((in_or_out == 'I') ? in_pix : out_pix),
		NULL);

	XtVaSetValues(w2,
		XmNbackgroundPixmap,
			((in_or_out != 'I') ? not_in_pix : not_out_pix),
		NULL);

	change_comment(forwho, "", FALSE);

	if (in_or_out == 'I') {
	    IOStat[forwho] &= ~OUT_FLAG;
	    IOStat[forwho] |= IN_FLAG;
	}
	else {
	    IOStat[forwho] &= ~IN_FLAG;
	    IOStat[forwho] |= OUT_FLAG;
	}

	if (andSend) sendToServer(STATUS_CHANGE, People[forwho]->Name, in_or_out);
}

void cbReasons(Widget w, caddr_t client, caddr_t)
{
    SelectedReason = (int) client;
    SelectedWidget = w;

    return;
}

void evMenuPopup(Widget w, caddr_t, XEvent *ev, char *)
{
    XButtonPressedEvent *but_event = (XButtonPressedEvent *) ev;

    XtUngrabPointer(w, CurrentTime);

    if (isButton1(but_event)) {
		XmMenuPosition(widget_list[COMMANDPOPUP], but_event);
		XtManageChild(widget_list[COMMANDPOPUP]);
    }
}

void cbSetGMsg(Widget, caddr_t client, caddr_t)
{
    if (*((char *) client) == 'R') {
		change_grpmsg('R', NULL, TRUE);
    }
    else {
		XmTextSetMaxLength(widget_list[GMSGFLD], MAXGRPMSG-1);
		XmTextSetString(widget_list[GMSGFLD], "");
		XtManageChild(gmpopup);
    }
    return;
}

void cbGMApply(Widget, caddr_t, caddr_t)
{
    char *txt = XmTextGetString(widget_list[GMSGFLD]);

    change_grpmsg('S', txt, TRUE);
    XtFree(txt);
 
    XtUnmanageChild(gmpopup);
}

void change_grpmsg(char mode, char *msg, bool andSend)
{
    Dimension GMh;

    if (mode == 'R' && !fGMShowing) return;
    else if (mode == 'R') {

		fGMShowing = fFortuneShowing = 0;
		XtUnmanageChild(widget_list[GROUPMSG]);

        restrictSize(MinWindowH, LeaveUnchanged, MinWindowH, LeaveUnchanged);

		XtVaSetValues(toplevel, XmNheight, MinWindowH, NULL);
    }

    if (mode == 'S') {
		XtVaSetValues(widget_list[GROUPMSG], XmNfontList, grpmsgfont, NULL);
		fFortuneShowing = 0;
    }
    else if (mode == 'F') {
		XtVaSetValues(widget_list[GROUPMSG], XmNfontList, fortunefont, NULL);
		fFortuneShowing = 1;
    }

    if (mode == 'S' || mode == 'F') {

		strncpy(GroupMsg, msg, MAXGRPMSG);
		GroupMsg[MAXGRPMSG] = '\0';

		XtUnmanageChild(widget_list[GROUPMSG]);
		setLabelString(GROUPMSG, GroupMsg);
		fGMShowing = 1;

		if (!XtIsManaged(widget_list[GROUPMSG])) 
		    XtManageChild(widget_list[GROUPMSG]);
		XtVaGetValues(widget_list[GROUPMSG], XmNheight, &GMh, NULL);

        restrictSize(MinWindowH+GMh,
			LeaveUnchanged, MinWindowH+GMh, LeaveUnchanged);

		XtVaSetValues(toplevel, XmNheight, (MinWindowH+GMh),NULL);

		XmUpdateDisplay(toplevel);
    }

	if (andSend) {
		MSG_TYPE m;
		if (mode == 'S') m = SET_GROUP_MSG;
		else if (mode == 'F') m = SET_FORTUNE;
		else if (mode == 'R') m = REMOVE_GROUP_MSG;

		sendToServer(m, "", 'x', (mode != 'R') ? GroupMsg : "");
	}
}

void cbAbout(Widget, caddr_t, caddr_t)
{
	static Widget aboutpopup = 0;
    static PopupButtonInfo aboutpbi[] = {
	{ "Dismiss", NULL, NULL },
	{ NULL }
    };
	Widget form, button;

	if (!aboutpopup) {

	    aboutpopup = createPopupWindow("About", toplevel, aboutpbi,
													&form,&button);

	    Widget w = XtVaCreateManagedWidget(
			VERSION, xmLabelWidgetClass, form,
			XmNleftAttachment, XmATTACH_FORM,
			XmNrightAttachment, XmATTACH_FORM,
    		XmNalignment, XmALIGNMENT_CENTER,
			NULL);

	    Widget rc = XtVaCreateManagedWidget("", xmRowColumnWidgetClass, form,
			XmNpacking, XmPACK_COLUMN,
			XmNtopAttachment, XmATTACH_WIDGET,
			XmNtopWidget, w,
			XmNtopOffset, 15,
			XmNbottomAttachment, XmATTACH_WIDGET,
			XmNbottomWidget, button,
			NULL);

		XtCreateManagedWidget(
			"Written by Shaun Terry (72010.1771@compuserve.com)",
			xmLabelWidgetClass, rc, NULL, 0);
		XtCreateManagedWidget(
			"Copyright (c) 1995, Freeware",
			xmLabelWidgetClass, rc, NULL, 0);
	}

	XtManageChild(aboutpopup);
}

static void fortune_on(Widget);
static void fortune_off(Widget);

static int fFortuneOn = True;

void change_fortune(char mode, char *msg, bool andSend)
{
    if ((fGMShowing && !fFortuneShowing) || !fFortuneOn) return;

    if (mode == 'R') change_grpmsg('R', NULL, andSend);
    else /*if (mode == 'S')*/ change_grpmsg('F', msg, andSend);
}

void cbFortune(Widget w, caddr_t, caddr_t)
{
    if (fFortuneOn) fortune_off(w);
    else fortune_on(w);
}

static void fortune_on(Widget w)
{
    setLabelString(w, "Fortune Off");
    fFortuneOn = True;
}

static void fortune_off(Widget w)
{
    setLabelString(w, "Fortune On");
    fFortuneOn = False;
	change_grpmsg('R', "", FALSE);
}

#define SECTIONEND	"EndSection"

//---------------------------------------- Constructor

ConfigFile::ConfigFile(char *name)
:fp(0), inReadSection(FALSE), fExists(FALSE)
{
	strcpy(fileName, name);
    if (open()) { fExists = TRUE; close(); } // Really should use stat()
}

//---------------------------------------- Open r or w

bool ConfigFile::open()
{
  fp = fopen(fileName, "r");
  return (!fp) ? FALSE : TRUE;
}

//---------------------------------------- Close

void ConfigFile::close() 
{ 
  if (fp) fclose(fp);
  fp = 0; 
}

//---------------------------------------- Position at beginning of a section

bool ConfigFile::seekSection(char *section)
{
    if (!open()) return FALSE;

    rewind(fp);

    char *p;
    while (p = nextLine()) {
	if (*p == '!') {
	    ++p;
	    skipWhiteSpace(p);
	    char *t = normalizeString(p);
	    if (!strncmp(t, section, strlen(section))) {
		inReadSection = TRUE;
		return TRUE;
	    }
	}
    }
 
    return FALSE;
}

char *ConfigFile::readNext()
{
    if (!fp) return 0;

    if (!inReadSection) {
	fprintf(stderr, "ConfigFile: Must call seekSection first!\n");
	return 0;
    }

    char *p = nextLine();

    if (!p) return 0;
    if (*p == '!') { inReadSection = FALSE; close(); return 0; }
    return normalizeString(p, FALSE);
}

char *ConfigFile::readField(char *fieldName)
{
    if (!open()) return 0;

    inReadSection = FALSE;

    rewind(fp);

    char *p;
    while (p = nextLine()) {
	if (*p > ' ' && *p <= '~') {
	    char *t = normalizeString(p);
	    if (!strncmp(t, fieldName, strlen(fieldName))) {
		close();
		while (*p != ':') {
		    if (*p == '\\') ++p;
		    ++p;
		}
		++p;
		skipWhiteSpace(p);
		return normalizeString(p);
	    }
	}
    }
    return 0;
}

char *ConfigFile::normalizeString(char *str, bool fColons)
{
    static char line[256];

    for (int i=0; *str; ++i, ++str) {

	if ((*str == ':' && fColons) || *str == '\n') break;
	if (*str == '\\') {
	    ++str;
	    if (*str == 'n') line[i] = '\n';
	    else if (*str == 't') line[i] = '\t';
	    else line[i] = *str;
	}
	else line[i] = *str;
    }
    line[i] = '\0';
    return line;
}

char *ConfigFile::denormalizeString(char *str, bool fColons)
{
    static char line[256];

    for (int i=0; *str; ++i, ++str) {

	if ((*str == ':' && fColons) || *str == '!') line[i++] = '\\';
	line[i] = *str;
    }
    line[i] = '\0';
    return line;
}

inline static void stripNewLineChar(char *p)
{
    register char *t;
    if (*(t = p+strlen(p)-1) == '\n') *t = '\0';
}

char *ConfigFile::nextLine()
{
    if (!fp) return 0;

    static char line[1024];

    while (1) {
	if (fgets(line, sizeof(line)-1, fp) == NULL) return 0;

	stripNewLineChar(line);

	// All lines ending in '\' should be strung together
	char *e = line + strlen(line) - 1;
	skipWhiteSpaceBackward(e, line);

	while (*e == '\\') {
	    char tmp[512];

	    if (fgets(tmp, sizeof(tmp)-1, fp) == NULL) break;
	    stripNewLineChar(tmp);
	    strcat(line+strlen(line)-1, tmp);

	    e = tmp + strlen(tmp) - 1;
	    skipWhiteSpaceBackward(e, tmp);
	}

	char *p = line;

	// Skip lines that are empty or comments
	skipWhiteSpace(p);
	if (*p == '\0' || *p == '#') continue;
	else break;
    }
    stripNewLineChar(line);
    return line;
}

char *ConfigFile::readField(char *fieldName, int *v)
{
    char *p = readField(fieldName);
    if (!p) return 0;

    *v = atoi(p);
    return p;
}

char *ConfigFile::readField(char *fieldName, long *v)
{
    char *p = readField(fieldName);
    if (!p) return 0;

    *v = atol(p);
    return p;
}

char *ConfigFile::readField(char *fieldName, double *v)
{
    char *p = readField(fieldName);
    if (!p) return 0;

    *v = atof(p);
    return p;
}

char *ConfigFile::readString(char *fieldName, char *dflt)
{
    char *p;
    if (!(p = readField(fieldName))) return dflt;
    return p;
}

int ConfigFile::readInt(char *fieldName, int dflt)
{
    int v;
    if (!readField(fieldName, &v)) return dflt;
    return v;
}

double ConfigFile::readFloat(char *fieldName, double dflt)
{
    double v;
    if (!readField(fieldName, &v)) return dflt;
    return v;
}

int connectToServer()
{
	struct sockaddr s;
	int l;
	// See if we're already connected
	if (server.sockFd() > 0) {
		if (getsockname(server.sockFd(), &s, &l) == 0) {
			if (fDebug) printf("Already connected\n");
			return 1;
		}
	}

	if (fDebug) printf("NOT connected\n");

	char *host = getenv("XINOUTHOST");
	if (host == NULL) host = "localhost";
	char *port = getenv("XINOUTPORT");
	int portNum = 0;
	if (port) portNum = atoi(port);

	// Try listening on the specified port
	if (portNum > 0) {
		if (server.openHost(host, portNum) != 0) {
			perror("connect");
			return(0);
		}
	}
	// Try listening on the service name
	else if (server.openHost(host, "xinout") == 0) {
		;	// OK
	}
	// Try listening on the default port name
	else if (server.openHost(host, 7232) == 0) {
		;	// OK
	}
	else { perror("connect: "); return 0; }

	if (inputId) XtRemoveInput(inputId);

	inputId = XtAppAddInput(_XtDefaultAppContext(),
					server.sockFd(),
					(XtPointer) XtInputReadMask, cbRead, NULL);

	sendToServer(INFO_REQUEST, "", 'x');

	return 1;
}

int sendToServer(MSG_TYPE type, char *who, char inout, char *text)
{
	if (!connectToServer()) return 0;

	inout_msg_hdr hdr;

	hdr.type = type;
	strcpy(hdr.who, who);
	hdr.in_or_out = inout;
	hdr.textlen = (strlen(text) > 0) ? strlen(text)+1 : 0;

	if (server.send((char *) &hdr, sizeof(hdr)) != sizeof(hdr)) {
		perror("send");
		return 0;
	}
	if (hdr.textlen > 0) server.send(text, hdr.textlen);

	return 1;
}

void cbRead(XtPointer, int *sock, XtInputId *id)
{
	if (fDebug) printf("Incoming message\n");
	inout_msg_hdr hdr;
	char *buf = 0;

	if (read(*sock, (char *) &hdr, sizeof(hdr)) <= 0) {
		server.close();
		XtRemoveInput(*id);
		inputId = 0;
		return;
	}
	if (hdr.textlen > 0) {
		buf = new char[hdr.textlen];
		read(*sock, buf, hdr.textlen);
	}

	if (fDebug) {
		printf("Type = %d  who = [%s] in_or_out = %c\n", hdr.type, hdr.who,
				hdr.in_or_out);
		printf("Text = [%s]\n", (buf) ? buf : "");
	}

    int who;

    if ((who = intwho(hdr.who)) < 0) {
		printf("xinout: inout_msg.who invalid (%s)\n", hdr.who);
		return;
    }

    if (hdr.type == STATUS_CHANGE)
		change_status(who, hdr.in_or_out, FALSE);
    else if (hdr.type == COMMENT_CHANGE)
		change_comment(who, (buf) ? buf : "", FALSE);
    else if (hdr.type == INFO_REQUEST) send_info();
    else if (hdr.type == REMOVE_GROUP_MSG) change_grpmsg('R', NULL, FALSE);
    else if (hdr.type == SET_GROUP_MSG) change_grpmsg('S', buf, FALSE);
    else if (hdr.type == REMOVE_FORTUNE) change_fortune('R', NULL, FALSE);
    else if (hdr.type == SET_FORTUNE) change_fortune('S', buf, FALSE);
    else printf("xinout: Invalid message type (%d)\n", hdr.type);

	if (buf) delete [] buf;
}

int intwho(char *who)
{
    int i;

    if (strlen(who) == 0) return 0;	/* Special case */

    for (i=0; i < NumPeople; ++i) {
		if (!strcmp(who, People[i]->Name)) return i;
    }
    return (-1);
}

void send_info()
{
    for (int i=0; i < NumPeople; ++i) {

		if (IOStat[i] & IN_FLAG)	/* Person is IN */
			sendToServer(STATUS_CHANGE, People[i]->Name, 'I');

		if (IOStat[i] & OUT_FLAG)	/* Person is OUT */
			sendToServer(STATUS_CHANGE, People[i]->Name, 'O');

		if (strlen(Comment[i]) > 0)
			sendToServer(COMMENT_CHANGE, People[i]->Name, 'x', Comment[i]);
    }

    if (fGMShowing && !fFortuneShowing)
		sendToServer(SET_GROUP_MSG, "", 'x', GroupMsg);
    else if (fFortuneShowing)
		sendToServer(SET_FORTUNE, "", 'x', GroupMsg);
}

void restrictSize(int minH, int minW, int maxH, int maxW)
{
    int mnH, mnW, mxH, mxW;
    Dimension d;

    if (!XtIsVendorShell(toplevel)) {
		fprintf(stderr, "Supplied Widget is not VendorShell.\n");
		return;
    }

    if (minH != LeaveUnchanged) {
	if (minH == SetToCurrent) {
		XtVaGetValues(toplevel, XmNheight, &d, NULL);
		mnH = (int) d;
	}
	else if (minH < 0) mnH = SetToUnrestricted;
		else mnH = minH;
		XtVaSetValues(toplevel, XmNminHeight, mnH, NULL);
    }

    if (minW != LeaveUnchanged) {
	if (minW == SetToCurrent) {
		XtVaGetValues(toplevel, XmNwidth, &d, NULL);
		mnW = (int) d;
	}
	else if (minW < 0) mnW = SetToUnrestricted;
	else mnW = minW;
	XtVaSetValues(toplevel, XmNminWidth, mnW, NULL);
    }

    if (maxH != LeaveUnchanged) {
	if (maxH == SetToCurrent) {
		XtVaGetValues(toplevel, XmNheight, &d, NULL);
		mxH = (int) d;
	}
	else if (maxH < 0) mxH = SetToUnrestricted;
	else mxH = maxH;
	XtVaSetValues(toplevel, XmNmaxHeight, mxH, NULL);
    }

    if (maxW != LeaveUnchanged) {
	if (maxW == SetToCurrent) {
		XtVaGetValues(toplevel, XmNwidth, &d, NULL);
		mxW = (int) d;
	}
	else if (maxW < 0) mxW = SetToUnrestricted;
	else mxW = maxW;
	XtVaSetValues(toplevel, XmNmaxWidth, mxW, NULL);
    }
}

#ifdef XXX
! XPM2  
64 64 4 1
  c #FFFFFFFFFFFF
. c #000000000000
X c #E5E5CBCB8989
o c #E5E52D2D2D2D

char *icon =
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"  ..........................................................    "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXoXoXXoXX.XXXoooXXoXXoXoooooXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXoXooXoXX.XXoXXXoXoXXoXXXoXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXoXoXooXX.XXoXXXoXoXXoXXXoXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXoXoXooXX.XXoXXXoXoXXoXXXoXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXoXoXXoXX.XXXoooXXXooXXXXoXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  ........................................................... . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .......................................................... .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  ........................................................... . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .......................................................... .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  ........................................................... . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .......................................................... .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  ........................................................... . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX. .  "
"  .XXXXXXXXXXXXXXXXXXXXXXXX.XXXXXXXXXX.XXXXXXXXXXXXXXXXXXXX.. . "
"  .......................................................... .  "
"      . . . . . . . . . . . . . . . . . . . . . . . . . . . . . "
"     . . . . . . . . . . . . . . . . . . . . . . . . . . . . .  "
"      . . . . . . . . . . . . . . . . . . . . . . . . . . . . . "
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"                                                                "
"                                                                ";
#endif

